theory Session1 imports Datatype begin

datatype 'a list = Nil
                 | Cons 'a "'a list"

text {* Example -- Lists *}
term "Nil"
term "Cons (0::nat) Nil"
term "Cons 0 (Cons 1 Nil)"

text {* Syntactic Sugar for Lists *}
notation Nil ("[]")
notation Cons (infixr "#" 65)

(* It would also be possible to use:

datatype 'a list = Nil ("[]")
                 | Cons 'a "'a list" (infixr "#" 65)
*)

text {* The notation is used before printing. *}
term "Nil"
term "Cons (0::nat) Nil"
term "Cons 0 (Cons 1 Nil)"

text {* Many lemmas for lists are proved automatically. *}
find_theorems (45) name:list

text {* Example -- Two Functions on Lists *}
primrec append :: "'a list => 'a list => 'a list" (infixr "@" 65)
where "[]@ys = ys"
    | "(x#xs)@ys = x#(xs@ys)"

primrec rev :: "'a list => 'a list"
where "rev [] = []"
    | "rev(x#xs) = rev xs @ (x#[])"

text {* An Introductory Proof *}
lemma append_Nil_right[simp]: "xs@[] = xs" by (induct xs) simp_all

lemma append_assoc[simp]: "(xs@ys)@zs = xs@(ys@zs)" by (induct xs) simp_all

lemma rev_append[simp]: "rev(xs@ys) = rev ys @ rev xs" by (induct xs) simp_all

theorem rev_rev_id[simp]: "rev(rev xs) = xs" by (induct xs) simp_all

text {* Definition -- Type Synonyms *}
types number   = nat
      gate     = "bool => bool => bool"
      'a plist = "('a * 'a)list"

typ gate
typ "'a plist"

text {* Definitions -- Constant Definitions *}
definition nand :: gate where "nand x y == ~(x & y)"
definition xor  :: gate where "xor  x y == (x & ~y) | (~x & y)"

thm nand_def
thm xor_def

text {* Exercises *}
text {* length *}
primrec length :: "'a list => nat"
where "length [] = 0"
    | "length(x#xs) = 1 + length xs"

lemma "length(xs@ys) = length xs + length ys" by (induct xs) simp_all

text {* snoc *}
primrec snoc :: "'a list => 'a => 'a list"
where "snoc [] y = y#[]"
    | "snoc (x#xs) y = x # snoc xs y"

lemma [simp]: "snoc (xs@ys) z = xs @ snoc ys z" by (induct xs) simp_all

lemma "rev(x#xs) = snoc(rev xs) x" by (induct xs) simp_all

text {* replace *}
primrec replace :: "'a => 'a => 'a list => 'a list"
where "replace x y [] = []"
    | "replace x y (z#zs) = (if x = z then y else z) # replace x y zs"

lemma [simp]: "replace w x (ys@zs) = replace w x ys @ replace w x zs" by (induct ys) simp_all

lemma "rev(replace x y zs) = replace x y (rev zs)" by (induct zs) simp_all

end
